@@ -1,15 +1,16 @@
-lmrob.lar <- function(x, y, control = lmrob.control(), mf = NULL)
+lmrob.lar <- function(x, y, control = lmrob.control(), ...)
 {
   ## LAR : Least Absolute Residuals -- i.e. L_1  M-estimate
   ## this function is identical to lmRob.lar of the robust package
 
+  ## '...': to be called as  'init(**, mf)' from lmrob()
   x <- as.matrix(x)
   p <- ncol(x)
   n <- nrow(x)
   stopifnot(p > 0, n >= p, length(y) == n, is.numeric(control$rel.tol))
   storage.mode(x) <- "double"
   storage.mode(y) <- "double"
-  bet0 <- 0.773372647623  ## bet0 = pnorm(0.75)
+  bet0 <- 0.773372647623  ## bet0 = pnorm(0.75); only for normalizing scale=SIGMA
   tmpn <- double(n)
   tmpp <- double(p)
 
@@ -31,8 +32,7 @@ lmrob.lar <- function(x, y, control = lmrob.control(), mf = NULL)
                  SC2=tmpp,
                  SC3=tmpp,
                  SC4=tmpp,
-                 BET0=as.double(bet0),
-                 PACKAGE = "robustbase")[c("THETA","SIGMA","RS","NIT","KODE")]
+                 BET0=as.double(bet0))[c("THETA","SIGMA","RS","NIT","KODE")]
   if (z1[5] > 1)
       stop("calculations stopped prematurely in rllarsbi\n",
            "(probably because of rounding errors).")
@@ -98,12 +98,10 @@ splitFrame <- function(mf, x = model.matrix(mt, mf),
 }
 
 ##' Compute M-S-estimator for linear regression ---> ../man/lmrob.M.S.Rd
-lmrob.M.S <- function(x, y, control, mf, split) {
-    if (missing(split))
-        split <- splitFrame(mf, x, control$split.type)
+lmrob.M.S <- function(x, y, control, mf, split = splitFrame(mf, x, control$split.type)) {
     if (ncol(split$x1) == 0) {
       warning("No categorical variables found in model. Reverting to S-estimator.")
-      return(lmrob.S(x, y, control, mf=mf))
+      return(lmrob.S(x, y, control))
     }
     if (ncol(split$x2) == 0) {
         warning("No continuous variables found in model. Reverting to L1-estimator.")
@@ -125,7 +123,7 @@ lmrob.M.S <- function(x, y, control, mf, split) {
     storage.mode(y) <- "double"
     c.chi <- .psi.conv.cc(control$psi, control$tuning.chi)
     traceLev <- as.integer(control$trace.lev)
-    z <- .C(R_lmrob_M_S,
+    z <- .C(R_lmrob_M_S, ## NB: If you change this, adapt ../inst/xtraR/m-s_fns.R
 	    x1,
 	    x2,
 	    y,
@@ -133,7 +131,7 @@ lmrob.M.S <- function(x, y, control, mf, split) {
             n=length(y),
             p1=ncol(x1),
             p2=ncol(x2),
-            nResample=as.integer(control$nResample),
+            nResample = as.integer(control$nResample),
             max_it_scale=as.integer(control$maxit.scale),
             scale=double(1),
             b1=double(ncol(x1)),
@@ -145,6 +143,7 @@ lmrob.M.S <- function(x, y, control, mf, split) {
             max_k=as.integer(control$k.max),
             rel_tol=as.double(control$rel.tol),
 	    inv_tol=as.double(control$solve.tol),
+	    scale_tol=as.double(control$scale.tol),
             converged = logical(1),
             trace_lev = traceLev,
             orthogonalize=TRUE,
@@ -155,7 +154,8 @@ lmrob.M.S <- function(x, y, control, mf, split) {
             )[c("b1","b2", "res","scale", "converged")]
 
     conv <- z$converged
-    ## coefficients
+    ## FIXME? warning if 'conv' is not ok ??
+    ## coefficients :
     idx <- split$x1.idx
     cf <- numeric(length(idx))
     cf[ idx] <- z$b1
